import unittest
import tempfile
import unittest
import numpy as np
import pandas as pd
import pyterrier_dr
from pyterrier_dr import FlexIndex

class TestFlexIndex(unittest.TestCase):

    def _generate_data(self, count=2000, dim=100):
        def random_unit_vec():
            v = np.random.rand(dim).astype(np.float32)
            return v / np.linalg.norm(v)
        return [
            {'docno': str(i), 'doc_vec': random_unit_vec()}
            for i in range(count)
        ]

    def test_compilation_with_rank_and_averageprf(self):
        self._test_compilation_with_rank_and_prf(pyterrier_dr.AveragePrf)

    def test_compilation_with_rank_and_vectorprf(self):
        self._test_compilation_with_rank_and_prf(pyterrier_dr.VectorPrf)
    
    def _test_compilation_with_rank_and_prf(self, prf_clz):

        with tempfile.TemporaryDirectory() as destdir:
            index = FlexIndex(destdir+'/index')
            dataset = self._generate_data(count=2000)
            index.index(dataset)
            
            retr = index.retriever()
            queries = pd.DataFrame([
                {'qid': '0', 'query_vec': dataset[0]['doc_vec']},
                {'qid': '1', 'query_vec': dataset[1]['doc_vec']},
            ])

            pipe1 = retr >> prf_clz(k=3) >> retr
            pipe1_opt = pipe1.compile()
            self.assertEqual(3, pipe1_opt[0].num_results)
            self.assertEqual(1000, pipe1_opt[-1].num_results)
            #NB: pipe1 wouldnt actually work for PRF, as doc_vecs are not present. however compilation is valid

            pipe2 = retr >> index.vec_loader() >> pyterrier_dr.AveragePrf(k=3) >> (retr % 2)
            pipe2_opt = pipe2.compile()
            self.assertEqual(3, pipe2_opt[0].num_results)
            self.assertEqual(2, pipe2_opt[-1].num_results)

            res2 = pipe2(queries)
            res2_opt = pipe2_opt(queries)
            
            pd.testing.assert_frame_equal(res2, res2_opt)
            
            


import pyterrier as pt
import pyterrier_dr
from . import FlexIndex


def _mmr(self, *, Lambda: float = 0.5, norm_rel: bool = False, norm_sim: bool = False, drop_doc_vec: bool = True, verbose: bool = False) -> pt.Transformer:
    """Returns an MMR (Maximal Marginal Relevance) scorer (i.e., re-ranker) over this index.

    The method first loads vectors from the index and then applies :class:`MmrScorer` to re-rank the results. See
    :class:`MmrScorer` for more details on MMR.

    Args:
        Lambda: The balance parameter between relevance and diversity (default: 0.5)
        norm_rel: Whether to normalize relevance scores to [0, 1] (default: False)
        norm_sim: Whether to normalize similarity scores to [0, 1] (default: False)
        drop_doc_vec: Whether to drop the 'doc_vec' column after re-ranking (default: True)
        verbose: Whether to display verbose output (e.g., progress bars) (default: False)

    .. cite.dblp:: conf/sigir/CarbonellG98
    """
    return self.vec_loader() >> pyterrier_dr.MmrScorer(Lambda=Lambda, norm_rel=norm_rel, norm_sim=norm_sim, drop_doc_vec=drop_doc_vec, verbose=verbose)
FlexIndex.mmr = _mmr



import unittest
import tempfile
import numpy as np
import pandas as pd
from pyterrier_dr import ils, FlexIndex


class TestIls(unittest.TestCase):
    def test_ils_basic(self):
        results = pd.DataFrame([
            ['q0', 'd0', np.array([0, 1, 0])],
            ['q0', 'd1', np.array([0, 1, 1])],
            ['q0', 'd2', np.array([1, 1, 0])],
            ['q0', 'd3', np.array([1, 1, 1])],
            ['q1', 'd0', np.array([0, 1, 0])],
            ['q2', 'd0', np.array([0, 1, 0])],
            ['q2', 'd1', np.array([0, 1, 1])],
        ], columns=['qid', 'docno', 'doc_vec'])
        result = dict(ils(results))
        self.assertAlmostEqual(result['q0'], 0.6874, places=3)
        self.assertAlmostEqual(result['q1'], 0.0000, places=3)
        self.assertAlmostEqual(result['q2'], 0.7071, places=3)

    def test_ils_vec_from_index(self):
        with tempfile.TemporaryDirectory() as d:
            index = FlexIndex(f'{d}/index.flex')
            index.index([
                {'docno': 'd0', 'doc_vec': np.array([0, 1, 0])},
                {'docno': 'd1', 'doc_vec': np.array([0, 1, 1])},
                {'docno': 'd2', 'doc_vec': np.array([1, 1, 0])},
                {'docno': 'd3', 'doc_vec': np.array([1, 1, 1])},
            ])
            results = pd.DataFrame([
                ['q0', 'd0'],
                ['q0', 'd1'],
                ['q0', 'd2'],
                ['q0', 'd3'],
                ['q1', 'd0'],
                ['q2', 'd0'],
                ['q2', 'd1'],
            ], columns=['qid', 'docno'])
            result = dict(ils(results, index))
            self.assertAlmostEqual(result['q0'], 0.6874, places=3)
            self.assertAlmostEqual(result['q1'], 0.0000, places=3)
            self.assertAlmostEqual(result['q2'], 0.7071, places=3)

    def test_ils_measure_from_index(self):
        with tempfile.TemporaryDirectory() as d:
            index = FlexIndex(f'{d}/index.flex')
            index.index([
                {'docno': 'd0', 'doc_vec': np.array([0, 1, 0])},
                {'docno': 'd1', 'doc_vec': np.array([0, 1, 1])},
                {'docno': 'd2', 'doc_vec': np.array([1, 1, 0])},
                {'docno': 'd3', 'doc_vec': np.array([1, 1, 1])},
            ])
            results = pd.DataFrame([
                ['q0', 'd0'],
                ['q0', 'd1'],
                ['q0', 'd2'],
                ['q0', 'd3'],
                ['q1', 'd0'],
                ['q2', 'd0'],
                ['q2', 'd1'],
            ], columns=['query_id', 'doc_id'])
            qrels = pd.DataFrame(columns=['query_id', 'doc_id', 'relevance']) # qrels ignored
            result = index.ILS.calc(qrels, results) 
            self.assertAlmostEqual(result.aggregated, 0.4648, places=3)
            self.assertEqual(3, len(result.per_query))
            self.assertEqual(result.per_query[0].query_id, 'q0')
            self.assertAlmostEqual(result.per_query[0].value, 0.6874, places=3)
            self.assertEqual(result.per_query[1].query_id, 'q1')
            self.assertAlmostEqual(result.per_query[1].value, 0.0000, places=3)
            self.assertEqual(result.per_query[2].query_id, 'q2')
            self.assertAlmostEqual(result.per_query[2].value, 0.7071, places=3)

    def test_ils_measure_from_index_cutoff(self):
        with tempfile.TemporaryDirectory() as d:
            index = FlexIndex(f'{d}/index.flex')
            index.index([
                {'docno': 'd0', 'doc_vec': np.array([0, 1, 0])},
                {'docno': 'd1', 'doc_vec': np.array([0, 1, 1])},
                {'docno': 'd2', 'doc_vec': np.array([1, 1, 0])},
                {'docno': 'd3', 'doc_vec': np.array([1, 1, 1])},
            ])
            results = pd.DataFrame([
                ['q0', 'd0'],
                ['q0', 'd1'],
                ['q0', 'd2'],
                ['q0', 'd3'],
                ['q1', 'd0'],
                ['q2', 'd0'],
                ['q2', 'd1'],
            ], columns=['query_id', 'doc_id'])
            qrels = pd.DataFrame(columns=['query_id', 'doc_id', 'relevance']) # qrels ignored
            result = (index.ILS@2).calc(qrels, results) 
            self.assertAlmostEqual(result.aggregated, 0.4714, places=3)
            self.assertEqual(3, len(result.per_query))
            self.assertEqual(result.per_query[0].query_id, 'q0')
            self.assertAlmostEqual(result.per_query[0].value, 0.7071, places=3)
            self.assertEqual(result.per_query[1].query_id, 'q1')
            self.assertAlmostEqual(result.per_query[1].value, 0.0000, places=3)
            self.assertEqual(result.per_query[2].query_id, 'q2')
            self.assertAlmostEqual(result.per_query[2].value, 0.7071, places=3)


if __name__ == '__main__':
    unittest.main()



import numpy as np
import torch
import pyterrier as pt
import pyterrier_alpha as pta

LIGHTNING_IR_AVAILIBLE = False
try:
    import lightning_ir as L
    LIGHTNING_IR_AVAILIBLE = True
except ImportError:
    pass


class LightningIRMonoScorer(pt.Transformer):
    def __init__(self,
                 model_name='webis/monoelectra-base',
                 batch_size=16,
                 text_field='text',
                 verbose=True,
                 device=None
                 ):
        if not LIGHTNING_IR_AVAILIBLE:
            raise ImportError("lightning_ir is required for LightningIRMonoScorer. Please install it via 'pip install lightning-ir'")
        self.model_name = model_name
        self.batch_size = batch_size
        self.text_field = text_field
        self.verbose = verbose
        if device is None:
            device = 'cuda' if torch.cuda.is_available() else 'cpu'
        self.device = torch.device(device)
        self.model = L.CrossEncoderModule(model_name).eval().to(self.device)

    def transform(self, inp):
        pta.validate.columns(inp, includes=['query', self.text_field])
        if len(inp) == 0:
            res = inp.assign(score=np.empty(shape=(0,), dtype=np.float32))
            pt.model.add_ranks(res)
            return res.sort_values(['qid', 'rank'])

        tmp = inp.reset_index().rename(columns={'index': '_row'})
        g = tmp.groupby('query', sort=False)[['_row', self.text_field]].agg(list)

        queries = g.index.tolist()                 # List[str]
        doclists = g[self.text_field].tolist()      # List[List[str]]
        idxlists = g['_row'].tolist()               # List[List[int]]

        out_scores = np.empty(len(inp), dtype=np.float32)

        n_groups = len(queries)
        iterator = range(n_groups)
        if self.verbose:
            iterator = pt.tqdm(iterator, total=n_groups, unit='query', desc=f'{self.model_name} scoring')

        start = 0
        for _ in iterator:
            if start >= n_groups:
                break

            # Pack queries until adding the next would exceed the pair budget
            pairs = 0
            end = start
            lengths = []
            while end < n_groups:
                need = len(doclists[end])
                if pairs > 0 and (pairs + need > self.batch_size):
                    break
                pairs += need
                lengths.append(need)
                end += 1

            q_batch = queries[start:end]
            d_batch = doclists[start:end]
            idx_batch = idxlists[start:end]

            with torch.no_grad(), torch.inference_mode():
                flat = self.model.score(q_batch, d_batch).scores.detach().cpu().numpy().reshape(-1)

            splits = np.cumsum(lengths)[:-1]
            per_query = np.split(flat, splits) if splits.size > 0 else [flat]
            for row_idxs, sc in zip(idx_batch, per_query):
                out_scores[np.asarray(row_idxs, dtype=int)] = sc

            start = end

        res = inp.assign(score=out_scores)
        pt.model.add_ranks(res)
        res = res.sort_values(['qid', 'rank'])
        return res



import numpy as np
import pandas as pd
import pyterrier as pt
import pyterrier_alpha as pta


class MmrScorer(pt.Transformer):
    schematic = {'label': 'MMR'}
    """An MMR (Maximal Marginal Relevance) scorer (i.e., re-ranker).

    The MMR scorer re-orders documents by balancing relevance (from the initial scores) and diversity (based on the
    similarity of the document vectors).

    .. cite.dblp:: conf/sigir/CarbonellG98
    """
    def __init__(self, *, Lambda: float = 0.5, norm_rel: bool = False, norm_sim: bool = False, drop_doc_vec: bool = True, verbose: bool = False):
        """
        Args:
            Lambda: The balance parameter between relevance and diversity (default: 0.5)
            norm_rel: Whether to normalize relevance scores to [0, 1] (default: False)
            norm_sim: Whether to normalize similarity scores to [0, 1] (default: False)
            drop_doc_vec: Whether to drop the 'doc_vec' column after re-ranking (default: True)
            verbose: Whether to display verbose output (e.g., progress bars) (default: False)
        """
        self.Lambda = Lambda
        self.norm_rel = norm_rel
        self.norm_sim = norm_sim
        self.drop_doc_vec = drop_doc_vec
        self.verbose = verbose

    def transform(self, inp: pd.DataFrame) -> pd.DataFrame:
        pta.validate.result_frame(inp, extra_columns=['doc_vec'])
        out = []

        if len(inp) == 0:
            return inp.assign(score=[], rank=[])

        it = inp.groupby('qid')
        if self.verbose:
            it = pt.tqdm(it, unit='q', desc=repr(self))

        for qid, frame in it:
            scores = frame['score'].values
            dvec_matrix = np.stack(frame['doc_vec'])
            dvec_matrix = dvec_matrix / np.linalg.norm(dvec_matrix, axis=1)[:, None]
            dvec_sims = dvec_matrix @ dvec_matrix.T
            if self.norm_rel:
                scores = (scores - scores.min()) / (scores.max() - scores.min())
            if self.norm_sim:
                dvec_sims = (dvec_sims - dvec_sims.min()) / (dvec_sims.max() - dvec_sims.min())
            marg_rels = np.zeros_like(scores)
            new_idxs = []
            for _ in range(scores.shape[0]):
                mmr_scores = (self.Lambda * scores) - ((1 - self.Lambda) * marg_rels)
                idx = mmr_scores.argmax()
                new_idxs.append(idx)
                if marg_rels.shape[0] > 1:
                    marg_rels = np.max(np.stack([marg_rels, dvec_sims[idx]]), axis=0)
                    marg_rels[idx] = float('inf') # ignore this document from now on
            new_frame = frame.iloc[new_idxs].reset_index(drop=True).assign(
                score=-np.arange(len(new_idxs), dtype=float),
                rank=np.arange(len(new_idxs))
            )
            if self.drop_doc_vec:
                new_frame = new_frame.drop(columns='doc_vec')
            out.append(new_frame)

        return pd.concat(out, ignore_index=True)

    __repr__ = pta.transformer_repr



from typing import Optional, Iterable, Tuple
import numpy as np
import pandas as pd
import ir_measures
import pyterrier as pt
from pyterrier_dr import FlexIndex


def ILS(index: FlexIndex, *, name: Optional[str] = None, verbose: bool = False) -> ir_measures.Measure: # noqa: N802
    """Create an ILS (Intra-List Similarity) measure calculated using the vectors in the provided index.

    Higher scores indicate lower diversity in the results.

    This measure supports the ``@k`` convention for applying a top-k cutoff before scoring.

    Args:
        index (FlexIndex): The index to use for loading document vectors.
        name (str, optional): The name of the measure (default: "ILS").
        verbose (bool, optional): Whether to display a progress bar.

    Returns:
        ir_measures.Measure: An ILS measure object.

    .. cite.dblp:: conf/www/ZieglerMKL05
    """
    return ir_measures.define(lambda qrels, results: _ils(results, index, verbose=verbose), name=name or 'ILS')


def ils(results: pd.DataFrame, index: Optional[FlexIndex] = None, *, verbose: bool = False) -> Iterable[Tuple[str, float]]:
    """Calculate the ILS (Intra-List Similarity) of a set of results.

    Higher scores indicate lower diversity in the results.

    Args:
        results: The result frame to calculate ILS for.
        index: The index to use for loading document vectors. Required if `results` does not have a `doc_vec` column.
        verbose: Whether to display a progress bar.

    Returns:
        Iterable[Tuple[str,float]]: An iterable of (qid, ILS) pairs.

    .. cite.dblp:: conf/www/ZieglerMKL05
    """
    return _ils(results.rename(columns={'docno': 'doc_id', 'qid': 'query_id'}), index, verbose=verbose)


def _ils(results: pd.DataFrame, index: Optional[FlexIndex] = None, *, verbose: bool = False) -> Iterable[Tuple[str, float]]:
    res = {}

    if index is not None:
        results = index.vec_loader()(results.rename(columns={'doc_id': 'docno'}))

    if 'doc_vec' not in results:
        raise ValueError('You must provide index to ils() if results do not have a `doc_vec` column.')

    it = results.groupby('query_id')
    if verbose:
        it = pt.tqdm(it, unit='q', desc='ILS')

    for qid, frame in it:
        if len(frame) > 1:
            vec_matrix = np.stack(frame['doc_vec'])
            vec_matrix = vec_matrix / np.linalg.norm(vec_matrix, axis=1)[:, None] # normalize vectors
            vec_sims = vec_matrix @ vec_matrix.T
            upper_right = np.triu_indices(vec_sims.shape[0], k=1)
            res[qid] = np.mean(vec_sims[upper_right])
        else:
            res[qid] = 0.0 # ILS is ill-defined when there's only one item.

    return res.items()
